#!/usr/bin/env ruby
require 'gtk2'
require 'fileutils'
require 'thread'                # Mutex

require 'editor'

# TODO win32: farby nevie prebrat z prikazu, undo/redo (novsi webkit?)
#             +bug s headingami
#             +uz nejde vobec editovat

# adresar = tema/poznamka, tam su aj prilozene subory

# korenovy adresar uzivatelovych poznamok
#ROOT = 'poznamky'
ROOT = File.expand_path('~/poznamky')
TRASH = '.trash'
NOTE = '__note__.html'
HELP = 'doc/help.html'

# store iter columns
NAME = 0
PATH = 1
TREE = 2
CHILD_PATH = 3 # used if '..', then refresh current path
ICON = 4


def open_file(tree, file)
  path = [tree.path, file].join('/')
  case Config::CONFIG['host_os']
  when /linux-gnu/i
    system("xdg-open '%s'" % path)
  when /mswin|windows/i
    system("start \"title\" \"%s\"" % path)
  when /darwin/i
    system("open '%s'" % path)
    # there can be more platforms with xdg-open
  when /./i
    system("xdg-open '%s'" % path)
  end
end

class Trash
  def initialize(root, view_trash, view_root)
    # dialog view trash and destination
    @view_trash = view_trash
    @view_root = view_root
    # root and trash tree
    @root = root
    @tree = Tree.new([ROOT, TRASH].join('/'), nil)

    refresh
  end

  def refresh(full=false)
    if full
      @tree = Tree.new([ROOT, TRASH].join('/'), nil)
    end
    @view_trash.model = @tree.treestore(nil)
    @view_root.model = @root.treestore(nil)
    @view_trash.expand_row(Gtk::TreePath.new('0'), false)
    @view_root.expand_row(Gtk::TreePath.new('0'), false)
  end

  def name(trashpath)
    pathsplit = trashpath.split('/')
    if pathsplit.length - 1 > [ROOT, TRASH].join('/').split('/').length
      return pathsplit[-1]
    else
      # it's in root, remove timestamp
      return pathsplit[-1].split('-')[0...-1].join('-')
    end
  end

  def restore(path, dest)
    oldname = name(path)
    oldpath = [dest, oldname].join('/')

    FileUtils.mv(path, oldpath)
    # remove from parent tree
    t = @tree.find(path).parent.remove(path)
    t.rename(oldpath)
    dtree = @root.find(dest)
    t.rebase(dtree)
    dtree.add(t)

    refresh
  end

  def delete(path)
    FileUtils.remove_entry_secure(path, force=false)
    @tree.remove(path)
    @view_trash.model = @tree.treestore(nil)
    @view_trash.expand_row(Gtk::TreePath.new('0'), false)
  end

  def self.add(path)
    name = [File.basename(path), '-', Time.now.to_i.to_s].join
    FileUtils.mv(path, [ROOT, TRASH, name].join('/'))
  end
end

# nacitanie adresarovej struktury
class Tree
   attr_reader :children, :path, :store, :parent

  def initialize(root, parent)
    @path = root
    @parent = parent
    @children = []
    @store = Gtk::ListStore.new(String, String, Tree, String, String)

    Dir["#{@path}/*"].each do |pathname|
      if FileTest.directory?(pathname)
        if pathname != @path
          @children.push(Tree.new(pathname, self))
        end
      end
    end

    @children.sort!
    fill
  end

  def search(name, phrase, tree)
    store = Gtk::ListStore.new(String)
    if phrase != ''
      re = Regexp.new(phrase)
    else
      re = nil
    end
    name = Regexp.new(name)
    tree.find_re(name, re, store)
    return store
  end

  def find_re(name, phrase, store)
    fn = [@path, NOTE].join('/')
    if @path != ROOT and File.basename(@path) =~ name and
        (phrase == nil or
         (File.exists?(fn) and
          File.open(fn) { |f| break true if f.grep(phrase) != [] }))
      iter = store.append
      iter[0] = @path.split('/')[ROOT.split('/').length .. -1].join('/')
    end

    @children.each { |c| c.find_re(name, phrase, store) }
  end

  def treestore(exclude)
    store = Gtk::TreeStore.new(String, String, Tree, Integer)
    iter = store.append(nil)
    iter[NAME] = '/'
    iter[PATH] = @path
    iter[TREE] = self
    fill_tree(store, iter, exclude)
    return store
  end

  def fill_tree(store, parent, exclude)
    @children.each { |c|
      next if c == exclude
      iter = store.append(parent)
      iter[NAME] = File.basename(c.path)
      iter[PATH] = c.path
      iter[TREE] = c

      c.fill_tree(store, iter, exclude)
    }
  end

  def rename(path)
    @path = path
    iter = @store.iter_first
    iter[NAME] = ".. (#{File.basename(@path)})"
    iter[CHILD_PATH] = @path

    if @parent
      parent = @parent
      @parent.remove(@path)
      parent.add(self)
    end
  end

  def parent=(p)
    @parent = p
  end

  def remove(path)
    @children.each_index { |i|
      if @children[i].path == path
        path = i
        path += 1 if self.path != ROOT and self.path != [ROOT, TRASH].join('/')
        @store.remove(@store.get_iter(path.to_s))
        @children[i].parent = nil
        return @children.delete_at(i)
      end
    }
    return nil
  end

  # ran in receiving tree
  def rebase(dest)
    @parent = dest
    @path = [dest.path, File.basename(@path)].join('/')

    @store.iter_first[PATH] = dest.path
    @store.iter_first[TREE] = @parent
    @store.iter_first[CHILD_PATH] = @path

    @store.each { |model, path, iter|
      if !iter[CHILD_PATH] # !".."
        iter[PATH] = [@path, iter[NAME]].join('/')
      end
    }

    @children.each { |c| c.rebase(self) }
  end

  def icon(path)
    if File.exists?([path, NOTE].join('/'))
      return 'gtk-yes' #Gtk::Stock::YES # NEW
    else
      return 'gtk-no' #Gtk::Stock::NO # MISSING_IMAGE / DIRECTORY
    end
  end

  def add(tree)
    tree.parent = self
    @children.each_index { |i|
      if @children[i].path > tree.path
        @children[i,0] = tree

        i += 1 if self.path != ROOT
        iter = @store.insert_before(@store.get_iter(i.to_s))
        iter[NAME] = File.basename(tree.path)
        iter[PATH] = tree.path
        iter[TREE] = tree
        iter[ICON] = icon(tree.path)
        # @store.each { |model, path, iter|
        #  if iter[PATH] > tree.path
        #    @store.insert_before(iter)
        #  end
        #}
        return
      end
    }
    # end of list
    @children.push(tree)
    iter = @store.append
    iter[NAME] = File.basename(tree.path)
    iter[PATH] = tree.path
    iter[TREE] = tree
    iter[ICON] = icon(tree.path)
  end

  def fill
    if @parent != nil
      iter = @store.append
      iter[NAME] = ".. (#{File.basename(@path)})"
      iter[PATH] = @path.split('/')[0...-1].join('/')
      iter[TREE] = @parent
      iter[CHILD_PATH] = @path

      iter[ICON] = Gtk::Stock::GO_BACK
    end

    @children.each { |c|
      iter = @store.append
      iter[NAME] = File.basename(c.path)
      iter[PATH] = c.path
      iter[TREE] = c
      iter[ICON] = icon(c.path)
    }
  end

  def find(path)
    if @path == path
      return self
    end

    @children.each { |c|
      ret = c.find(path)
      return ret if ret
    }

    return nil
  end

  # porovnanie prvkov pre sort
  def <=>(a)
    @path <=> a.path
  end
end

class PathButton < Gtk::ToggleButton
  attr_reader :image

  def initialize(name, icon)
    super()

    @image = Gtk::Image.new(icon, Gtk::IconSize::BUTTON)
    @label = Gtk::Label.new(name)

    h = Gtk::HBox.new
    h.pack_start(@image)
    h.pack_start(@label, false, false, 2)
    add(h)
  end

  def text
    @label.text
  end

  def text=(str)
    @label.text = str
  end
end

class PathButtons < Gtk::HBox
  def initialize(win)
    super()
    @path = '/'
    @buttons = []
    @window = win

    @button_handler = Proc.new { |b|
      if b == @active # keep activated
        if !b.active?
          b.active = true
        end
      else
        # it's the activated button?
        if b.active?
          # deactivate currently active
          tmp = @active
          @active = b
          tmp.active = false if tmp

          path = [ROOT].concat(self.subpath(@buttons.index(b))).join('/')

          if !@init
            child = @buttons[@buttons.index(b) + 1]
            if child
              up = [path, child.text].join('/')
              iter = @window.tree.find(up).store.iter_first
              @window.chdir(iter, @move)
            else
              tree = @window.tree.find(path)
              tree.parent.store.each { |model, path, iter|
                if iter[TREE] == tree
                  @window.chdir(iter, @move)
                  break
                end
              }
            end

            @move = false
          end
        end
      end
    }

    @init = true
    button(@path, 'gtk-home')
    @init = false
  end

  def button(name, icon)
    button = PathButton.new(name, icon)
    button.signal_connect('toggled', &@button_handler)
    @buttons << button
    button.active = true

    self.pack_start(button, false, false, 1)
  end

  def go(filepath, move)
    # move cursor? true for row activation
    @move = move

    path = filepath.split('/')[ROOT.split('/').length .. -1].join('/')
    path = ['/'].concat(path.split('/'))

    pairs = path.zip(@buttons)
    pairs.each_index { |i|
      if !pairs[i][1] or pairs[i][0] != pairs[i][1].text
        self.children[i..-1].each { |b|
          self.remove(b)
          @buttons.delete(b)
        }

        path[i..-1].each_index { |j|
          if File.exists?([ROOT].concat(path[1..i]).concat([NOTE]).join('/'))
            icon = 'gtk-file'
          end
          button(path[i+j], icon)
        }

        break
      end
    }
    @buttons[path.length - 1].activate
    self.show_all
  end

  def subpath(index)
    path = []
    @buttons[1, index].each { |b|
      path << b.text
    }
    return path
  end

  def path_match(oldpath)
    oldpath = oldpath.split('/')[ROOT.split('/').length .. -1]
    path = self.subpath(oldpath.length)

    if path == oldpath
      return oldpath.length
    else
      return nil
    end
  end

  def on_save(oldpath)
    len = path_match(oldpath)
    if len
      @buttons[len].image.stock = 'gtk-file'
    end
  end

  def on_rename(oldpath, name)
    len = path_match(oldpath)
    if len
      @buttons[len].text = name
    end
  end

  def on_remove(oldpath)
    len = path_match(oldpath)
    if len
      @buttons[len .. -1].each { |b|
        self.remove(b)
        @buttons.delete(b)
      }
    end
  end
end


# hlavne okno - strom poznamok
class MainWindow
  attr_accessor :tree, :path

  def initialize
    @restore_open = Mutex.new
    @move_open = Mutex.new

    @editors = []
    @window = Gtk::Window.new(Gtk::Window::TOPLEVEL)
    @window.resizable = true
    @window.title = "Notes"
    @window.set_default_size(700, 500)

    # nacitanie struktury
    @tree = Tree.new(ROOT, nil)

    # pridanie treeview
    @view = Gtk::TreeView.new(@tree.store)
    @view.width_request = 200
    @view.height_request = 300

    @renderer = Gtk::CellRendererText.new
    @renderer.signal_connect('edited') { |renderer, path, text|
      @renderer.editable = false
      iter = @view.model.get_iter(path)
      path = iter[PATH].split('/')[0...-1].push(text).join('/')

      @path.on_rename(iter[PATH], text)

      if not File.exist?(path)
        File.rename(iter[PATH], path)
        iter[TREE].rename(path)
      end

      @trash.refresh(true) if @trash
    }
    @renderer.signal_connect('editing-canceled') { |renderer, entry, path|
      @renderer.editable = false
    }
    # @renderer.signal_connect('editing-started') { |renderer, entry, path|
    # }

    # Add column using our renderer
    # @col = Gtk::TreeViewColumn.new("Notes", @renderer, :text => NAME)
    @col = Gtk::TreeViewColumn.new("Notes")
    renderer = Gtk::CellRendererPixbuf.new
    @col.pack_start(renderer, false)
    @col.add_attribute(renderer, :stock_id, ICON)
    @col.pack_start(@renderer, true)
    @col.add_attribute(@renderer, :text, NAME)
    @view.append_column(@col)

    # scrollovatelne treeview
    @treescr = Gtk::ScrolledWindow.new
    @treescr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    @treescr.add(@view)

    # attachments
    store = Gtk::ListStore.new(String)
    @attach = Gtk::TreeView.new(store)
    # @renderer = Gtk::CellRendererText.new
    # Add column using our renderer
    @col2 = Gtk::TreeViewColumn.new("Attachments", @renderer, :text => 0)
    @attach.append_column(@col2)

    # scrollovatelne treeview
    @attachscr = Gtk::ScrolledWindow.new
    @attachscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    @attachscr.add(@attach)

    @attach.signal_connect('row-activated') {
      iter = @attach.selection.selected
      open_file(@preview, iter[0])
    }

    gc = Gtk::RC.get_style(@window).bg(Gtk::STATE_NORMAL).to_s
    color = "##{gc[1...3]}#{gc[5...7]}#{gc[9...11]}"
    @blank_page = %Q{<body bgcolor="#{color}"></body>}
    @webview = Gtk::WebKit::WebView.new
    @webview.load_html_string(@blank_page, '')

    webviewscr = Gtk::ScrolledWindow.new
    webviewscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
    webviewscr.add(@webview)

    @root = @tree
    @view.signal_connect('row-activated') { |view, path, column|
      row = view.selection.selected
      filepath = row[PATH]

      # move cursor only when going through '..'
      if row[CHILD_PATH]
        @path.go(filepath, true)
      else
        @path.go(filepath, false)
      end
    }

    vpan = Gtk::VPaned.new
    vpan.add1(@treescr)
    vpan.add2(@attachscr)

    @previewlabel = Gtk::Label.new
    # @previewlabel.set_alignment(0,0)
    @path = PathButtons.new(self)

    preview = Gtk::VBox.new
    preview.pack_start(@path, false)
    preview.pack_start(@previewlabel, false)
    preview.pack_start(webviewscr)

    hpan = Gtk::HPaned.new
    hpan.add1(vpan)
    hpan.add2(preview)

    @ui = generate_ui
    @window.add_accel_group(@ui.accel_group)
    @menubar = @ui.get_widget("/menubar_main")

    @commands = Gtk::HBox.new
    [['Rename (F2)', @on_rename], ['Preview (F3)', @on_view],
     ['Open (F4)', @on_open], ['Move (F6)', @on_move],
     ['Create (F7)', @on_create], ['Delete (F8)', @on_delete]].each { |name|
      button = Gtk::Button.new(name[0])
      button.signal_connect('clicked') { name[1].call }
      @commands.pack_start(button)
    }

    vbox = Gtk::VBox.new
    vbox.pack_start(@menubar, false, false)
    vbox.pack_start(hpan)
    vbox.pack_start(@commands, false, false)

    @window.add(vbox)
    @window.signal_connect('delete_event', &@on_quit)
    @window.show_all
  end

  def chdir(row, move)
    @view.model = row[TREE].store
    storepath = nil

    if move or !row[TREE].parent
      if row[CHILD_PATH] # ".."
        @view.model.each { |model, path, iter|
          if row[CHILD_PATH] == iter[TREE].path
            storepath = path
            break
          end
        }
      else
        storepath = @view.model.iter_first.path
      end

      @view.set_cursor(storepath, nil, nil)
    else
      @view.set_cursor(Gtk::TreePath.new('0'), nil, nil)
    end

    @root = row[TREE]
    filepath = row[PATH]
    fn = [filepath, NOTE] .join('/')
    # don't preview on row-activate: NOT ANYMORE
    refresh(fn, row, true)
  end

  # filename; row iter; activated row or just preview
  def refresh(fn, row, activated)
    if File.exists?(fn) and (!activated or row[PATH] != ROOT)
      File.open(fn) { |fd|
        @webview.load_html_string(fd.read(), "file://%s" % File.expand_path(fn))
      }
    else
      @webview.load_html_string(@blank_page, '')
    end

    @attach.model.clear
    path = fn.split('/')[0...-1].join('/')
    if path != ROOT
      files = []
      Dir["#{path}/*"].each do |pathname|
        basename = File.basename(pathname)
        if not FileTest.directory?(pathname) and basename != NOTE
          files << basename
        end
      end
      files.sort!.each { |fn| @attach.model.append[0] = fn }

      @previewlabel.markup= "Preview of <b>%s</b>" % path.split('/')[ROOT.split('/').length..-1].join(' / ')
    else
      @previewlabel.text = ''
    end
  end

  def get_path(iter)
    if iter and iter[CHILD_PATH] # ".."
      iter[CHILD_PATH]
    else
      iter[PATH]
    end
  end

  def open?(fn, tree)
    if tree
      fn = fn.split('/')
      @editors.each { |e|
        if fn == e.filename.split('/')[0...fn.length]
          return true
        end
      }
    else
      @editors.each { |e|
        if fn == e.filename
          return true
        end
      }
    end
    return false
  end

  def generate_ui
    @on_quit = Proc.new do |actiongroup, action|
      if @editors.length > 0
        dialog = Gtk::MessageDialog.new(@window,
                                        Gtk::Dialog::DESTROY_WITH_PARENT,
                                        Gtk::MessageDialog::WARNING,
                                        Gtk::MessageDialog::BUTTONS_OK_CANCEL,
                                        'Editor windows are open')
        dialog.secondary_text = 'Are you sure you want to quit?'
        dialog.signal_connect('response') { |dialog, response|
          if response == Gtk::Dialog::RESPONSE_OK
            Gtk.main_quit
          end
          dialog.destroy
        }
        dialog.show_all
      else
        Gtk.main_quit
      end
    end

    @on_open = Proc.new do |actiongroup, action|
      iter = @view.selection.selected

      if iter and (path = get_path(iter)) != ROOT
        file = [path, NOTE].join('/')
        # allow only one editor window per file
        if not open?(file, false)
          if iter[CHILD_PATH]
            tree = iter[TREE].children.each { |c|
              if c.path == iter[CHILD_PATH]
                  break c
              end
            }
          else
            tree = iter[TREE]
          end

          e = Editor.new("%s [TakeNote]" % file.split('/')[ROOT.split('/').length...-1].join('/'),
                         file, @editors, tree, self)
          @editors.push(e)
        end
      end
    end

    @on_view = Proc.new do |actiongroup, actiong|
      path = @view.cursor[0]
      if path
        row = @view.model.get_iter(path)

        @preview = row[TREE]
        path = get_path(row)
        fn = [path, NOTE].join('/')
        refresh(fn, row, false)
      end
    end

    on_about = Proc.new do |actiongroup, actiong|
      Gtk::AboutDialog.show(@window,
                            { :program_name => 'TakeNote',
                              :version => '1.0 beta',
                              :copyright => 'Copyright © 2009 Adam Sloboda, Veronika Wolfová' })
    end

    on_help = Proc.new do |actiongroup, actiong|
      fn = HELP
      if File.exists?(fn)
        @previewlabel.markup = '<b>Help</b>'
        File.open(fn) { |fd|
          @webview.load_html_string(fd.read(),
                                    "file://%s" % File.expand_path(fn))
        }
      else
        @previewlabel.markup = '<b>Help not found!</b>'
        @webview.load_html_string(@blank_page, '')
      end
    end

    @on_rename = Proc.new do |actiongroup, actiong|
      row = @view.selection.selected
      if row and !row[CHILD_PATH] # !".."
        fn = [row[PATH], NOTE].join('/')
        if open?(fn, false)
          ErrorBox.new(@window, 'Error', "This note is open in the editor.")
        else
          @renderer.editable = true
          path = @view.selection.selected.path
          @view.set_cursor(path, @col, true)
        end
      end
    end

    @on_move = Proc.new do |actiongroup, actiong|
      root = @root
      row = @view.selection.selected
      if row and !row[CHILD_PATH] and @move_open.try_lock # not ".."
        if open?(row[PATH], true)
          ErrorBox.new(@window, 'Error', "Note from this tree is open in the editor.")
          @move_open.unlock
        else
          treeview = Gtk::TreeView.new(@tree.treestore(row[TREE]))
          treeview.headers_visible = false
          treeview.expand_row(Gtk::TreePath.new('0'), false)
          renderer = Gtk::CellRendererText.new
          col = Gtk::TreeViewColumn.new("Note tree", @renderer, :text => NAME)
          treeview.append_column(col)
          treescr = Gtk::ScrolledWindow.new
          treescr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
          treescr.add(treeview)

          dialog = Gtk::Dialog.new("Move note",
                                   @window,
                                   Gtk::Dialog::DESTROY_WITH_PARENT,
                                   [ Gtk::Stock::OK, Gtk::Dialog::RESPONSE_OK ],
                                   [ Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL ])
          dialog.signal_connect('response') { |dialog, response|
            iter = treeview.selection.selected
            if response == Gtk::Dialog::RESPONSE_OK and iter
              if File.exists?([iter[PATH], row[NAME]].join('/'))
                ErrorBox.new(@window, "Error", "There is already \"%s\" in destination directory." % row[NAME])
              else
                # errors here
                FileUtils.mv(row[PATH], iter[PATH])
                @path.on_remove(row[PATH])
                tree = root.remove(row[PATH])
                tree.rebase(iter[TREE])
                iter[TREE].add(tree)

                @trash.refresh(true) if @trash
              end
            end
            if not (!iter and response == Gtk::Dialog::RESPONSE_OK)
              dialog.destroy
              @move_open.unlock
            end
          }

          treeview.signal_connect('row-activated') {
            dialog.response(Gtk::Dialog::RESPONSE_OK)
          }

          dialog.vbox.pack_start(Gtk::Label.new("Choose destination for \"%s\":" % row[NAME]),
                                 false, false)
          dialog.vbox.pack_start(treescr)
          dialog.set_default_size(250, 350)
          dialog.show_all
        end
      end
    end

    @on_create = Proc.new do |actiongroup, actiong|
      root = @root

      num = 1
      name = "New Note (#{num})"
      root.children.each { |c|
        p c.path.split('/')[-1]
        if name == c.path.split('/')[-1]
          num = num + 1
          name = "New Note (#{num})"
        end
      }

      path = [root.path, name].join('/')
      FileUtils.mkdir(path)
      root.add(Tree.new(path, root))

      root.store.each { |model, path, iter|
        if iter[NAME] == name
          @renderer.editable = true
          path = iter.path
          @view.set_cursor(path, @col, true)
        end
      }
    end

    @on_delete = Proc.new do |actiongroup, actiong|
      row = @view.selection.selected
      root = @root

      if row and !row[CHILD_PATH] # not ".."
        if open?(row[PATH], true)
          ErrorBox.new(@window, 'Error', "Note from this tree is open in the editor.")
        else
          dialog = Gtk::MessageDialog.new(@window,
                                          Gtk::Dialog::DESTROY_WITH_PARENT,
                                          Gtk::MessageDialog::QUESTION,
                                          Gtk::MessageDialog::BUTTONS_YES_NO,
                                          'Delete')
          dialog.secondary_text = "Are you sure you want to delete \"%s\"?" % row[NAME]
          dialog.signal_connect('response') { |dialog, response|
            if response == Gtk::Dialog::RESPONSE_YES
              iter = @view.selection.selected

              Trash.add(iter[PATH])
              @path.on_remove(iter[PATH])

              path = row[PATH]
              refresh(path, nil, false)
              root.remove(iter[PATH])

              @trash.refresh(true) if @trash
            end
            dialog.destroy
          }
          dialog.show_all
        end
      end
    end

    on_restore = Proc.new do |actiongroup, actiong|
      if @restore_open.try_lock
        renderer = Gtk::CellRendererText.new
        trashview = Gtk::TreeView.new
        col = Gtk::TreeViewColumn.new("Trash", @renderer, :text => 0)
        trashview.append_column(col)
        trashscr = Gtk::ScrolledWindow.new
        trashscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
        trashscr.add(trashview)
        trashscr.set_size_request(250, -1)

        rootview = Gtk::TreeView.new
        col = Gtk::TreeViewColumn.new("Destination", @renderer, :text => 0)
        rootview.append_column(col)
        rootscr = Gtk::ScrolledWindow.new
        rootscr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
        rootscr.add(rootview)
        rootscr.set_size_request(250, -1)

        hpan = Gtk::HPaned.new
        hpan.add1(trashscr)
        hpan.add2(rootscr)

        @trash = Trash.new(@tree, trashview, rootview)

        dialog = Gtk::Dialog.new("Restore",
                                 @window,
                                 Gtk::Dialog::DESTROY_WITH_PARENT,
                                 [ 'Res_tore', 0 ],
                                 [ Gtk::Stock::OPEN, 4 ],
                                 [ Gtk::Stock::DELETE, 1 ],
                                 [ Gtk::Stock::CLOSE, 3])
        dialog.set_default_size(500, 400)
        dialog.signal_connect('response') { |dialog, response|
          titer = trashview.selection.selected
          riter = rootview.selection.selected
          case response
          when 0 # restore
            if titer and titer != trashview.model.iter_first
              if not riter
                # default restore point is root
                riter = rootview.model.iter_first
              end
              name = @trash.name(titer[PATH])
              if File.exists?([riter[PATH], name].join('/'))
                ErrorBox.new(@window, "Error", "There is already \"%s\" in destination directory." % [name])
              else
                @trash.restore(titer[PATH], riter[PATH])
              end
            end
            
          when 1 # delete
            if titer and titer != trashview.model.iter_first
              ddialog = Gtk::MessageDialog.new(@window,
                                               Gtk::Dialog::DESTROY_WITH_PARENT,
                                               Gtk::MessageDialog::QUESTION,
                                               Gtk::MessageDialog::BUTTONS_YES_NO,
                                               'Delete')
              ddialog.secondary_text = "Are you sure you want to delete \"%s\"?" % titer[NAME]
              ddialog.signal_connect('response') { |dialog, response|
                if response == Gtk::Dialog::RESPONSE_YES
                  @trash.delete(titer[PATH])
                end
                ddialog.destroy
              }
              ddialog.show_all
            end
          when 3 # close
            dialog.destroy
            @restore_open.unlock
            @trash = nil
          when 4 # open
            iter = trashview.selection.selected
            if iter and iter[PATH] != [ROOT, TRASH].join('/')
              trashview.row_activated(iter.path, nil)
            end
          else                  # delete event
            @restore_open.unlock
            @trash = nil
          end
        }

        trashview.signal_connect('row-activated') {
          row = trashview.selection.selected

          filepath = [row[PATH], NOTE].join('/')
          if File.exists?(filepath)
            editor = Gtk::WebKit::WebView.new
            editor.open(File.expand_path(filepath))

            scroll = Gtk::ScrolledWindow.new
            scroll.add(editor)
            scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
            
            w = Gtk::Window.new(row[NAME])
            w.set_default_size(550, 400)
            w.add(scroll)
            w.show_all
          else
            ErrorBox.new(dialog, 'Error', "There is no saved note in \"%s\"." % row[NAME])
          end
        }
        
        dialog.vbox.add(hpan)
        dialog.show_all
      end
    end

    on_search = Proc.new do |actiongroup, actiong|
      table = Gtk::Table.new(2, 2)
      label = Gtk::Label.new('Search name:')
      input_name = Gtk::Entry.new
      table.attach(label,      0, 1, 0, 1, xopt=Gtk::SHRINK)
      table.attach(input_name, 1, 2, 0, 1, xopt=Gtk::FILL)
      h2 = Gtk::HBox.new
      label = Gtk::Label.new('Search phrase:')
      input_phrase = Gtk::Entry.new
      table.attach(label,        0, 1, 1, 2, xopt=Gtk::SHRINK)
      table.attach(input_phrase, 1, 2, 1, 2, xopt=Gtk::FILL)
      view = Gtk::TreeView.new
      view.headers_visible = false
      renderer = Gtk::CellRendererText.new
      col = Gtk::TreeViewColumn.new("Note", @renderer, :text => 0)
      view.append_column(col)
      treescr = Gtk::ScrolledWindow.new
      treescr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
      treescr.add(view)

      # search in subtree
      ftreeview = Gtk::TreeView.new(@tree.treestore(nil))
      ftreeview.headers_visible = false
      ftreeview.expand_row(Gtk::TreePath.new('0'), false)
      renderer = Gtk::CellRendererText.new
      col = Gtk::TreeViewColumn.new("Note tree", @renderer, :text => NAME)
      ftreeview.append_column(col)
      ftreescr = Gtk::ScrolledWindow.new
      ftreescr.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)
      ftreescr.add(ftreeview)

      # select current tree
      ftreeview.model.each { |model, path, iter|
        if iter[PATH] == @root.path
          ftreeview.set_cursor(path, nil, nil)
        end
      }

      h = Gtk::HPaned.new
      h.pack1(ftreescr, true, false)
      h.pack2(treescr, true, false)

      dialog = Gtk::Dialog.new("Search",
                               @window,
                               Gtk::Dialog::DESTROY_WITH_PARENT,
                               [ Gtk::Stock::FIND, 0 ],
                               [ Gtk::Stock::CLOSE, 1])
      dialog.set_default_size(400, 300)
      dialog.signal_connect('response') { |dialog, response|
        if response == 0
          if ftreeview.selection.selected
            tree = ftreeview.selection.selected[TREE]
          else
            tree = @tree
          end
          view.model = @tree.search(input_name.text, input_phrase.text, tree)
        else
          dialog.destroy
        end
      }

      view.signal_connect('row-activated') { |view, path, column|
        row = view.selection.selected
        filepath = [ROOT, row[0]].join('/')
        @path.go(filepath, false)

        # need different row type
        #refresh([filepath, NOTE].join('/'), row, true)

        # nechavame otvorene
        # dialog.response(1)
      }

      dialog.vbox.pack_start(table, false, false)
      #dialog.vbox.add(treescr)
      dialog.vbox.add(h)
      dialog.show_all
    end

    ui_def = <<END
    <ui>
      <menubar name="menubar_main">
        <menu action="menuFile">
          <menuitem action="open" />
          <menuitem action="search" />
          <menuitem action="refresh" />
          <menuitem action="restore" />
          <separator />
          <menuitem action="quit" />
        </menu>
        <menu action="menuCommands">
          <!-- <menuitem action="help" /> -->
          <menuitem action="rename" />
          <menuitem action="view" />
          <menuitem action="edit" />
          <menuitem action="move" />
          <menuitem action="create" />
          <menuitem action="delete" />
        </menu>
        <menu action="menuHelp">
          <menuitem action="help" />
          <menuitem action="about" />
        </menu>
      </menubar>
    </ui>
END

# TODO html search globalny
#on_tmp = Proc.new {
# @editor.search_text (s, case_sensitive, forward, wrap) -> b
# @editor.mark_text_matches (s, case_sensitive, limit) -> i

#p @webview.search_text('t', false, true, true)
#n = @webview.mark_text_matches('t', false, 0)
# @webview.highlight_text_matches = true
# @webview.unmark_text_matches(self)
#}

    actions = Gtk::ActionGroup.new("Actions")
    actions.add_actions([
      ["menuFile", nil, "_File"],
      ["menuCommands", nil, "_Commands"],
      ["menuHelp", nil, "_Help"],

      ["open", Gtk::Stock::OPEN, "_Open", "<ctrl>O", nil, @on_open],
      ["refresh", Gtk::Stock::REFRESH, "_Refresh preview", "<ctrl>R", nil, @on_view],
      ["search", Gtk::Stock::FIND, "_Search", "<ctrl>F", nil, on_search],
      ["restore", 'gnome-stock-trash', "_Trash", nil, nil, on_restore],
      ["quit", Gtk::Stock::QUIT, "_Quit", "<ctrl>Q", nil, @on_quit],

      ["rename", nil, "_Rename", "F2", nil, @on_rename],
      ["view", nil, "_Preview", "F3", nil, @on_view],
      ["edit", nil, "_Open", "F4", nil, @on_open],
      ["move", nil, "_Move", "F6", nil, @on_move],
      ["create", nil, "_Create", "F7", nil, @on_create],
      ["delete", nil, "_Delete", "F8", nil, @on_delete],

      ["help", Gtk::Stock::HELP, "_Help", "F1", nil, on_help],
      ["about", Gtk::Stock::ABOUT, "_About", nil, nil, on_about]
])

    ui = Gtk::UIManager.new
    ui.insert_action_group(actions, 0)
    ui.add_ui(ui_def)

    return ui
  end
end


class ErrorBox < Gtk::MessageDialog
  def initialize(parent, header, msg)
    super(parent, Gtk::Dialog::DESTROY_WITH_PARENT,
          Gtk::MessageDialog::ERROR,
          Gtk::MessageDialog::BUTTONS_OK,
          header)
    self.secondary_text = msg
    self.signal_connect('response') {
      self.destroy
    }
    self.show_all
  end
end



if not File.exists?(ROOT)
  FileUtils.mkdir(ROOT)
end

if not File.exists?([ROOT, TRASH].join('/'))
  FileUtils.mkdir([ROOT, TRASH].join('/'))
end

main = MainWindow.new
Gtk.main
